"""Environment object for compiler.

This module contains the various column accessors and function evaluators that
are made available by the query compiler via their compilation context objects.
Define new columns and functions here.
"""
__copyright__ = "Copyright (C) 2014-2017  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import decimal
import re
import textwrap

from decimal import Decimal

import dateutil.parser
from dateutil.relativedelta import relativedelta, weekday

from beancount.core.number import ZERO
from beancount.core import amount
from beancount.core import position
from beancount.core import inventory
from beancount.core import account
from beancount.core import convert
from beancount.core import prices
from beancount.core.account_types import get_account_sign, get_account_sort_key

from beanquery import query_compile
from beanquery import types


class ColumnsRegistry(dict):

    def register(self, dtype):
        def decorator(func):
            class Col(query_compile.EvalColumn):
                def __init__(self):
                    super().__init__(dtype)
                __call__ = staticmethod(func)
            Col.__name__ = func.__name__
            Col.__doc__ = func.__doc__
            self[Col.__name__] = Col()
            return func
        return decorator


def function(intypes, outtype, pass_context=None, name=None):
    def decorator(func):
        class Func(query_compile.EvalFunction):
            __intypes__ = intypes
            pure = not pass_context
            def __init__(self, context, operands):
                super().__init__(context, operands, outtype)
            def __call__(self, row):
                args = [operand(row) for operand in self.operands]
                for arg in args:
                    if arg is None:
                        return None
                if pass_context:
                    return func(self.context, *args)
                return func(*args)
        Func.__name__ = name if name is not None else func.__name__
        Func.__doc__ = func.__doc__
        query_compile.FUNCTIONS[Func.__name__].append(Func)
        return func
    return decorator


def register(name=None):
    def decorator(cls):
        if name is not None:
            cls.__name__ = name
        query_compile.FUNCTIONS[cls.__name__].append(cls)
        return cls
    return decorator


@register('getitem')
class GetItem2(query_compile.EvalFunction):
    __intypes__ = [dict, str]

    def __init__(self, context, operands):
        super().__init__(context, operands, object)

    def __call__(self, row):
        obj, key = self.operands
        obj = obj(row)
        if obj is None:
            return None
        return obj.get(key(row))


@register('getitem')
class GetItem3(query_compile.EvalFunction):
    __intypes__ = [dict, str, types.Any]

    def __init__(self, context, operands):
        super().__init__(context, operands, object)

    def __call__(self, row):
        obj, key, default = self.operands
        obj = obj(row)
        if obj is None:
            return None
        return obj.get(key(row), default(row))


def Function(name, args):
    func = types.function_lookup(query_compile.FUNCTIONS, name, args)
    if func is not None:
        return func(None, args)
    raise KeyError


## Type casting


@function([types.Any], bool, name='bool')
def bool_(x):
    """Convert to bool value."""
    return bool(x)


@function([int], int, name='int')
@function([bool], int, name='int')
@function([Decimal], int, name='int')
@function([str], int, name='int')
@function([object], int, name='int')
def int_(x):
    try:
        return int(x)
    except (ValueError, TypeError):
        return None


@function([Decimal], Decimal, name='decimal')
@function([int], Decimal, name='decimal')
@function([bool], Decimal, name='decimal')
@function([str], Decimal, name='decimal')
@function([object], Decimal, name='decimal')
def decimal_(x):
    try:
        return Decimal(x)
    except (ValueError, TypeError, decimal.InvalidOperation):
        return None


@function([types.Any], str, name='str')
def str_(x):
    if x is True:
        return 'TRUE'
    if x is False:
        return 'FALSE'
    return str(x)


@function([datetime.date], datetime.date, name='date')
@function([str], datetime.date, name='date')
@function([object], datetime.date, name='date')
def date_(x):
    if isinstance(x, datetime.date):
        return x
    if isinstance(x, str):
        try:
            return datetime.datetime.strptime(x, '%Y-%m-%d').date()
        except ValueError:
            pass
    return None


@function([int, int, int], datetime.date, name='date')
def date_from_ymd(year, month, day):
    """Construct a date with year, month, day arguments."""
    try:
        return datetime.date(year, month, day)
    except ValueError:
        return None


## Functions


@function([Decimal], Decimal)
@function([amount.Amount], amount.Amount)
@function([position.Position], position.Position)
@function([inventory.Inventory], inventory.Inventory)
def neg(x):
    """Negative value."""
    return -x


@function([Decimal], Decimal, name='abs')
@function([position.Position], position.Position, name='abs')
@function([inventory.Inventory], inventory.Inventory, name='abs')
def abs_(x):
    """Absolute value."""
    return abs(x)


@function([Decimal, Decimal], Decimal)
@function([Decimal, int], Decimal)
def safediv(x, y):
    """A division operation that traps division by zero exceptions and outputs zero instead."""
    if y == 0:
        return ZERO
    return x / y


@function([Decimal], Decimal, name='round')
@function([Decimal, int], Decimal, name='round')
@function([int], int, name='round')
@function([int, int], int, name='round')
def round_(num, digits=0):
    """Round the argument."""
    return round(num, digits)


@function([list], int)
@function([set], int)
@function([str], int)
def length(x):
    """Compute the length of the argument. This works on sequences."""
    return len(x)


@function([types.Any], str, name='repr')
def repr_(x):
    """Convert the argument to a string via repr()."""
    return repr(x)


@function([str, int], str)
def maxwidth(x, n):
    """Convert the argument to a substring. This can be used to ensure
    maximum width. This will insert ellipsis ([...]) if necessary."""
    return textwrap.shorten(x, width=n)


@function([str, int, int], str)
def substr(string, start, end):
    """Extract a substring of the argument."""
    return string[start:end]


@function([str, str, int], str)
def splitcomp(string, delim, index):
    """Split a string and extract one of its components."""
    return string.split(delim)[index]


# Operations on dates.

@function([datetime.date], int)
def year(x):
    """Extract the year from a date."""
    return x.year


@function([datetime.date], int)
def month(x):
    """Extract the month from a date."""
    return x.month


@function([datetime.date], int)
def day(x):
    """Extract the day from a date."""
    return x.day


@function([datetime.date], datetime.date)
def yearmonth(x):
    """Extract the year and month from a date."""
    return datetime.date(x.year, x.month, 1)


@function([datetime.date], str)
def quarter(x):
    """Extract the quarter from a date."""
    return '{:04d}-Q{:1d}'.format(x.year, (x.month - 1) // 3 + 1)


@function([datetime.date], str, name='weekday')
def weekday_(x):
    """Extract a 3-letter weekday from a date."""
    return x.strftime('%a')


@function([], datetime.date)
def today():
    """Today's date"""
    return datetime.date.today()


# Operations on accounts.

@function([str], str)
@function([str, int], str)
def root(acc, n=1):
    """Get the root name(s) of the account."""
    return account.root(n, acc)


@function([str], str)
def parent(acc):
    """Get the parent name of the account."""
    return account.parent(acc)


@function([str], str)
def leaf(acc):
    """Get the name of the leaf subaccount."""
    return account.leaf(acc)


@function([str, str], str)
def grep(pattern, string):
    """Match a regular expression against a string and return only the matched portion."""
    match = re.search(pattern, string)
    if match:
        return match.group(0)
    return None


@function([str, str, int], str)
def grepn(pattern, string, n):
    """Match a pattern with subgroups against a string and return the subgroup at the index."""
    match = re.search(pattern, string)
    if match:
        return match.group(n)
    return None


@function([str, str, str], str)
def subst(pattern, repl, string):
    """Substitute leftmost non-overlapping occurrences of pattern by replacement."""
    return re.sub(pattern, repl, string)


@function([str], str)
def upper(string):
    """Convert string to uppercase."""
    return string.upper()


@function([str], str)
def lower(string):
    """Convert string to lowercase."""
    return string.lower()


# Avoid building a tuple for each function invocation.
NONENONE = None, None


@function([str], datetime.date, pass_context=True)
def open_date(context, acc):
    """Get the date of the open directive of the account."""
    open_entry, _ = context.tables['accounts'].accounts.get(acc, NONENONE)
    if open_entry is None:
        return None
    return open_entry.date


@function([str], datetime.date, pass_context=True)
def close_date(context, acc):
    """Get the date of the close directive of the account."""
    _, close_entry = context.tables['accounts'].accounts.get(acc, NONENONE)
    if close_entry is None:
        return None
    return close_entry.date


@function([str], dict, pass_context=True)
@function([str, str], object, pass_context=True)
def open_meta(context, account, key=None):
    """Get the metadata dict of the open directive of the account."""
    open_entry, _ = context.tables['accounts'].accounts.get(account, NONENONE)
    if open_entry is None:
        return None
    if key is None:
        return open_entry.meta
    return open_entry.meta.get(key)


# Stub kept only for function type checking and for generating documentation.
@function([str], object)
def meta(context, key):
    """Get some metadata key of the posting."""
    raise NotImplementedError


# Stub kept only for function type checking and for generating documentation.
@function([str], object)
def entry_meta(context, key):
    """Get some metadata key of the transaction."""
    raise NotImplementedError


# Stub kept only for function type checking and for generating documentation.
@function([str], object)
def any_meta(context, key):
    """Get metadata from the posting or its parent transaction if not present."""
    raise NotImplementedError


@function([str], dict, pass_context=True)
@function([str, str], object, pass_context=True)
@function([str], dict, pass_context=True, name='commodity_meta')
@function([str, str], object, pass_context=True, name='commodity_meta')
def currency_meta(context, commodity, key=None):
    """Get the metadata dict of the commodity directive of the currency."""
    entry = context.tables['commodities'].commodities.get(commodity)
    if entry is None:
        return None
    if key is None:
        return entry.meta
    return entry.meta.get(key)


@function([str], str, pass_context=True)
def account_sortkey(context, acc):
    """Get a string to sort accounts in order taking into account the types."""
    account_types = context.tables['accounts'].types
    index, name = get_account_sort_key(account_types, acc)
    return '{}-{}'.format(index, name)


# Stub kept only for function type checking and for generating documentation.
@function([str], bool)
def has_account(context, pattern):
    """True if the transaction has at least one posting matching the regular expression argument."""
    raise NotImplementedError


# Note: Don't provide this, because polymorphic multiplication on Amount,
# Position, Inventory isn't supported yet.
#
# class AccountSign(query_compile.EvalFunction):
#     "Produce a +1 / -1 signed value to multiply with to correct balances."
#     __intypes__ = [str]
#
#     def __init__(self, operands):
#         super().__init__(operands, Decimal)
#
#     def __call__(self, context):
#         args = self.eval_args(context)
#         return Decimal(account_types.get_account_sign(args[0], context.account_types))

# Operation on inventories, positions and amounts.


@function([position.Position], amount.Amount, name='units')
def position_units(pos):
    """Get the number of units of a position (stripping cost)."""
    return convert.get_units(pos)


@function([inventory.Inventory], inventory.Inventory, name='units')
def inventory_units(inv):
    """Get the number of units of an inventory (stripping cost)."""
    return inv.reduce(convert.get_units)


@function([position.Position], amount.Amount, name='cost')
def position_cost(pos):
    """Get the cost of a position."""
    return convert.get_cost(pos)


@function([inventory.Inventory], inventory.Inventory, name='cost')
def inventory_cost(inv):
    """Get the cost of an inventory."""
    return inv.reduce(convert.get_cost)


@function([amount.Amount, str], amount.Amount, pass_context=True, name='convert')
@function([amount.Amount, str, datetime.date], amount.Amount, pass_context=True, name='convert')
def convert_amount(context, amount_, currency, date=None):
    """Coerce an amount to a particular currency."""
    price_map = context.tables['prices'].price_map
    return convert.convert_amount(amount_, currency, price_map, date)


@function([position.Position, str], amount.Amount, pass_context=True, name='convert')
@function([position.Position, str, datetime.date], amount.Amount, pass_context=True, name='convert')
def convert_position(context, pos, currency, date=None):
    """Coerce an amount to a particular currency."""
    price_map = context.tables['prices'].price_map
    return convert.convert_position(pos, currency, price_map, date)


@function([inventory.Inventory, str], inventory.Inventory, pass_context=True, name='convert')
@function([inventory.Inventory, str, datetime.date], inventory.Inventory, pass_context=True, name='convert')
def convert_inventory(context, inv, currency, date=None):
    """Coerce an inventory to a particular currency."""
    price_map = context.tables['prices'].price_map
    return inv.reduce(convert.convert_position, currency, price_map, date)


@function([position.Position], amount.Amount, pass_context=True, name='value')
@function([position.Position, datetime.date], amount.Amount, pass_context=True, name='value')
def position_value(context, pos, date=None):
    """Convert a position to its cost currency at the market value."""
    price_map = context.tables['prices'].price_map
    return convert.get_value(pos, price_map, date)


@function([inventory.Inventory], inventory.Inventory, pass_context=True, name='value')
@function([inventory.Inventory, datetime.date], inventory.Inventory, pass_context=True, name='value')
def inventory_value(context, inv, date=None):
    """Coerce an inventory to its market value."""
    price_map = context.tables['prices'].price_map
    return inv.reduce(convert.get_value, price_map, date)


@function([str, str], Decimal, pass_context=True)
@function([str, str, datetime.date], Decimal, pass_context=True, name='getprice')
def getprice(context, base, quote, date=None):
    """Fetch a price."""
    price_map = context.tables['prices'].price_map
    pair = (base.upper(), quote.upper())
    _, price = prices.get_price(price_map, pair, date)
    return price


@function([amount.Amount], Decimal)
def number(x):
    """Extract the number from an Amount."""
    return x.number


@function([amount.Amount], str)
@function([amount.Amount], str, name='commodity')
def currency(x):
    """Extract the currency from an Amount."""
    return x.currency


@function([str, set], str)
def findfirst(pattern, values):
    """Filter a string sequence by regular expression and return the first match."""
    if not values:
        return None
    for value in sorted(values):
        if re.match(pattern, value):
            return value
    return None


@function([set], str)
def joinstr(values):
    """Join a sequence of strings to a single comma-separated string."""
    return ','.join(values)


@function([str, inventory.Inventory], amount.Amount, name='only')
def only_inventory(currency, inventory_):
    """Get one currency's amount from the inventory."""
    return inventory_.get_currency_units(currency)


@function([inventory.Inventory], bool, name='empty')
def empty_inventory(inventory_):
    """Determine whether the inventiry is empty."""
    return inventory_.is_empty()


@function([position.Position, str], position.Position, name='filter_currency')
def filter_currency_position(pos, currency):
    """Filter an inventory to just the specified currency."""
    return pos if pos.units.currency == currency else None


@function([inventory.Inventory, str], inventory.Inventory, name='filter_currency')
def filter_currency_inventory(inv, currency):
    """Filter an inventory to just the specified currency."""
    return inventory.Inventory(pos for pos in inv if pos.units.currency == currency)


@function([Decimal, str], Decimal, pass_context=True)
@function([amount.Amount, str], amount.Amount, pass_context=True)
@function([position.Position, str], position.Position, pass_context=True)
@function([inventory.Inventory, str], inventory.Inventory, pass_context=True)
def possign(context, x, account):
    """Correct sign of an Amount based on the usual balance of associated account."""
    account_types = context.tables['accounts'].types
    sign = get_account_sign(account, account_types)
    return x if sign >= 0  else -x


# ``date`` type

class Date(types.Structure):
    name = 'date'
    columns = ColumnsRegistry()

    @columns.register(int)
    def year(x):
        return x.year

    @columns.register(int)
    def month(x):
        return x.month

    @columns.register(int)
    def day(x):
        return x.day

types.ALIASES[datetime.date] = Date


@function([str], datetime.date)
@function([str, str], datetime.date)
def parse_date(string, frmt=None):
    """Parse date from string."""
    if frmt is None:
        return dateutil.parser.parse(string).date()
    return datetime.datetime.strptime(string, frmt).date()


@function([datetime.date, datetime.date], int)
def date_diff(x, y):
    """Calculates the difference (in days) between two dates."""
    return (x - y).days


@function([datetime.date, int], datetime.date)
def date_add(x, y):
    """Adds/subtracts number of days from the given date."""
    return x + datetime.timedelta(days=y)


@function([str, datetime.date], datetime.date)
def date_trunc(field, x):
    """Truncate a date to the specified precision."""
    if field == 'week':
        return x - relativedelta(weekday=weekday(0, -1))
    if field == 'month':
        return datetime.date(x.year, x.month, 1)
    if field == 'quarter':
        return datetime.date(x.year, x.month - (x.month - 1) % 3, 1)
    if field == 'year':
        return datetime.date(x.year, 1, 1)
    if field == 'decade':
        return datetime.date(x.year - x.year % 10, 1, 1)
    if field == 'century':
        return datetime.date(x.year - (x.year - 1) % 100, 1, 1)
    if field == 'millennium':
        return datetime.date(x.year - (x.year - 1) % 1000, 1, 1)
    return None


@function([str, datetime.date], int)
def date_part(field, x):
    """Extract the specified field from a date."""
    if field == 'weekday' or field == 'dow':
        return x.weekday()
    if field == 'isoweekday' or field == 'isodow':
        return x.isoweekday()
    if field == 'week':
        # isocalendar() returns a named tuple only in Python >= 3.9.
        return x.isocalendar()[1]
    if field == 'month':
        return x.month
    if field == 'quarter':
        return (x.month - 1) // 3 + 1
    if field == 'year':
        return x.year
    if field == 'isoyear':
        # isocalendar() returns a named tuple only in Python >= 3.9.
        return x.isocalendar()[0]
    if field == 'decade':
        return x.year // 10
    if field == 'century':
        return (x.year - 1) // 100 + 1
    if field == 'millennium':
        return (x.year - 1) // 1000 + 1
    if field == 'epoch':
        return int((x - datetime.date(1970, 1, 1)).total_seconds())
    return None


@function([str], relativedelta)
def interval(x):
    """Construct a relative time interval."""
    m = re.fullmatch(r'([-+]?[0-9]+)\s+(day|month|year)s?', x)
    if not m:
        return None
    number = int(m.group(1))
    unit = m.group(2)
    if unit == 'day':
        return relativedelta(days=number)
    if unit == 'week':
        return relativedelta(weeks=number)
    if unit == 'month':
        return relativedelta(months=number)
    if unit == 'year':
        return relativedelta(years=number)
    if unit == 'decade':
        return relativedelta(years=number * 10)
    if unit == 'century':
        return relativedelta(years=number * 100)
    if unit == 'millennium':
        return relativedelta(years=number * 1000)
    return None


@function([relativedelta, datetime.date, datetime.date], datetime.date)
def date_bin(stride, source, origin):
    """Bin a date into the specified stride aligned with the specified origin.

    As an extension to the the SQL standard ``date_bin()`` function this
    function also accepts strides containing units of months and years.
    """
    if stride.months or stride.years:
        if origin + stride <= origin:
            # FIXME: this should raise and error: stride must be greater than zero
            return None
        if source >= origin:
            d = n = origin
            while True:
                n += stride
                if n >= source:
                    return d
                d = n
        else:
            n = origin
            while True:
                n -= stride
                if n <= source:
                    return n
    else:
        seconds = stride.days * 86400 + stride.hours * 3600 + stride.minutes * 60 + stride.seconds
        if seconds < 0:
            # FIXME: this should raise and error: stride must be greater than zero
            return None
        diff = (source - origin).total_seconds()
        modulo = diff % seconds
        delta = diff - modulo
        result = origin + datetime.timedelta(seconds=delta)
        if modulo < 0:
            result -= datetime.timedelta(seconds=seconds)
        return result


@function([str, datetime.date, datetime.date], datetime.date, name='date_bin')
def date_bin_str(stride, source, origin):
    return date_bin(interval(stride), source, origin)


def aggregator(intypes, name=None):
    def decorator(cls):
        cls.__intypes__ = intypes
        if name is not None:
            cls.__name__ = name
        query_compile.FUNCTIONS[cls.__name__].append(cls)
        return cls
    return decorator


@aggregator([types.Asterisk], name='count')
class Count(query_compile.EvalAggregator):
    """Count the number of input rows."""
    def __init__(self, context, operands):
        super().__init__(context, operands, int)

    def update(self, store, context):
        store[self.handle] += 1


@aggregator([types.Any], name='count')
class CountArg(query_compile.EvalAggregator):
    """Count the number of non-NULL occurrences of the argument."""
    def __init__(self, context, operands):
        super().__init__(context, operands, int)

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle] += 1


@aggregator([int], name='sum')
class SumInt(query_compile.EvalAggregator):
    """Calculate the sum of the numerical argument."""
    def __init__(self, context, operands):
        super().__init__(context, operands, operands[0].dtype)

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle] += value


@aggregator([Decimal], name='sum')
class SumDecimal(query_compile.EvalAggregator):
    """Calculate the sum of the numerical argument."""
    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle] += value


@aggregator([amount.Amount], name='sum')
class SumAmount(query_compile.EvalAggregator):
    """Calculate the sum of the amount. The result is an Inventory."""
    def __init__(self, context, operands):
        super().__init__(context, operands, inventory.Inventory)

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle].add_amount(value)


@aggregator([position.Position], name='sum')
class SumPosition(query_compile.EvalAggregator):
    """Calculate the sum of the position. The result is an Inventory."""
    def __init__(self, context, operands):
        super().__init__(context, operands, inventory.Inventory)

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle].add_position(value)


@aggregator([inventory.Inventory], name='sum')
class SumInventory(query_compile.EvalAggregator):
    """Calculate the sum of the inventories. The result is an Inventory."""
    def __init__(self, context, operands):
        super().__init__(context, operands, inventory.Inventory)

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            store[self.handle].add_inventory(value)


@aggregator([types.Any], name='first')
class First(query_compile.EvalAggregator):
    """Keep the first of the values seen."""
    def initialize(self, store):
        store[self.handle] = None

    def update(self, store, context):
        if store[self.handle] is None:
            value = self.operands[0](context)
            store[self.handle] = value


@aggregator([types.Any], name='last')
class Last(query_compile.EvalAggregator):
    """Keep the last of the values seen."""
    def initialize(self, store):
        store[self.handle] = None

    def update(self, store, context):
        value = self.operands[0](context)
        store[self.handle] = value


@aggregator([types.Any], name='min')
class Min(query_compile.EvalAggregator):
    """Compute the minimum of the values."""
    def initialize(self, store):
        store[self.handle] = None

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            cur = store[self.handle]
            if cur is None or value < cur:
                store[self.handle] = value


@aggregator([types.Any], name='max')
class Max(query_compile.EvalAggregator):
    """Compute the maximum of the values."""
    def initialize(self, store):
        store[self.handle] = None

    def update(self, store, context):
        value = self.operands[0](context)
        if value is not None:
            cur = store[self.handle]
            if cur is None or value > cur:
                store[self.handle] = value



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import functools
import re
import sys
import textwrap
import unittest

import click.testing

from beancount import loader
from beancount.utils import test_utils

from beanquery import shell


@functools.lru_cache(None)
def load():
    entries, errors, options = loader.load_string(textwrap.dedent("""
      2022-01-01 open Assets:Checking         USD
      2022-01-01 open Assets:Federal:401k     IRAUSD
      2022-01-01 open Assets:Gold             GLD
      2022-01-01 open Assets:Vacation         VACHR
      2022-01-01 open Assets:Vanguard:RGAGX   RGAGX
      2022-01-01 open Expenses:Commissions    USD
      2022-01-01 open Expenses:Food           USD
      2022-01-01 open Expenses:Home:Rent      USD
      2022-01-01 open Expenses:Taxes:401k     IRAUSD
      2022-01-01 open Expenses:Taxes:Federal  USD
      2022-01-01 open Expenses:Tests          USD
      2022-01-01 open Expenses:Vacation       VACHR
      2022-01-01 open Income:ACME             USD
      2022-01-01 open Income:Gains            USD
      2022-01-01 open Income:Vacation         VACHR

      2022-01-01 * "ACME" "Salary"
        Assets:Checking           10.00 USD
        Income:ACME              -11.00 USD
        Expenses:Taxes:Federal     1.00 USD
        Assets:Federal:401k       -2.00 IRAUSD
        Expenses:Taxes:401k        2.00 IRAUSD
        Assets:Vacation               5 VACHR
        Income:Vacation              -5 VACHR

      2022-01-01 * "Rent"
        Assets:Checking           42.00 USD
        Expenses:Home:Rent        42.00 USD

      2022-01-02 * "Holidays"
        Assets:Vacation              -1 VACHR
        Expenses:Vacation

      2022-01-03 * "Test 01"
        Assets:Checking            1.00 USD
        Expenses:Tests

      2022-01-04 * "My Fovorite Plase" "Eating out alone"
        Assets:Checking            4.00 USD
        Expenses:Food

      2022-01-05 * "Invest"
        Assets:Checking         -359.94 USD
        Assets:Vanguard:RGAGX     2.086 RGAGX {172.55 USD}

      2013-10-23 * "Buy Gold"
        Assets:Checking        -1278.67 USD
        Assets:Gold                   9 GLD {141.08 USD}
        Expenses:Commissions          8.95 USD

      2022-01-07 * "Sell Gold"
        Assets:Gold                 -16 GLD {147.01 USD} @ 135.50 USD
        Assets:Checking         2159.05 USD
        Expenses:Commissions       8.95 USD
        Income:Gains             184.16 USD

      2022-01-08 * "Sell Gold"
        Assets:Gold                 -16 GLD {147.01 USD} @ 135.50 USD
        Assets:Checking         2159.05 USD
        Expenses:Commissions       8.95 USD
        Income:Gains             184.16 USD

      2022-02-01 * "ACME" "Salary"
        Assets:Checking           10.00 USD
        Income:ACME              -11.00 USD
        Expenses:Taxes:Federal     1.00 USD
        Assets:Federal:401k       -2.00 IRAUSD
        Expenses:Taxes:401k        2.00 IRAUSD
        Assets:Vacation               5 VACHR
        Income:Vacation              -5 VACHR

      2022-02-01 * "Rent"
        Assets:Checking           43.00 USD
        Expenses:Home:Rent        43.00 USD

      2022-02-02 * "Test 02"
        Assets:Checking            2.00 USD
        Expenses:Tests

      2030-01-01 query "taxes" "
        SELECT
          date, description, position, balance
        WHERE
          account ~ 'Taxes'
        ORDER BY date DESC
        LIMIT 20"

      2015-01-01 query "home" "
        SELECT
          last(date) as latest,
          account,
          sum(position) as total
        WHERE
          account ~ ':Home:'
        GROUP BY account"

    """))
    return entries, errors, options


def run_shell_command(cmd):
    """Run a shell command and return its output."""
    with test_utils.capture('stdout') as stdout, test_utils.capture('stderr') as stderr:
        shell_obj = shell.BQLShell('', sys.stdout)
        entries, errors, options = load()
        shell_obj.context.attach('beancount:', entries=entries, errors=errors, options=options)
        shell_obj._extract_queries(entries)  # pylint: disable=protected-access
        shell_obj.onecmd(cmd)
    return stdout.getvalue(), stderr.getvalue()


def runshell(function):
    """Decorate a function to run the shell and return the output."""
    def wrapper(self):
        out, err = run_shell_command(function.__doc__)
        return function(self, out, err)
    return wrapper


class TestUseCases(unittest.TestCase):
    """Testing all the use cases from the proposal here.
    I'm hoping to replace reports by these queries instead."""

    @runshell
    def test_print_from(self, out, err):
        """
        PRINT FROM narration ~ 'alone'
        """
        self.assertRegex(out, 'Eating out alone')

    @runshell
    def test_accounts(self, out, err):
        """
        SELECT DISTINCT account, open_date(account)
        ORDER BY account_sortkey(account);
        """
        self.assertRegex(out, 'Assets:Checking *2022-01-01')
        self.assertRegex(out, 'Income:ACME *2022-01-01')

    @runshell
    def test_commodities(self, out, err):
        """
        SELECT DISTINCT currency ORDER BY 1;
        """
        self.assertRegex(out, 'USD')
        self.assertRegex(out, 'IRAUSD')
        self.assertRegex(out, 'VACHR')

    @runshell
    def test_commodities_cost(self, out, err):
        """
        SELECT DISTINCT cost_currency ORDER BY 1;
        """
        self.assertRegex(out, 'USD')

    @runshell
    def test_commodities_pairs(self, out, err):
        """
        SELECT DISTINCT currency, cost_currency ORDER BY 1, 2;
        """
        self.assertRegex(out, 'GLD *USD')

    @runshell
    def test_balances(self, out, err):
        """
        BALANCES AT cost;
        """
        self.assertRegex(out, r'Assets:Gold *\d+\.\d+ USD')

    @runshell
    def test_balances_with_where(self, out, err):
        """
        JOURNAL 'Assets:Checking';
        """
        self.assertRegex(out, 'Salary')

    @runshell
    def test_balance_sheet(self, out, err):
        """
        BALANCES AT cost
        FROM OPEN ON 2022-01-02 CLOSE ON 2022-02-01 CLEAR;
        """
        self.assertRegex(out, r'Assets:Gold * \d+\.\d+ USD')

    @runshell
    def test_income_statement(self, out, err):
        """
        SELECT account, cost(sum(position))
        FROM OPEN ON 2022-01-01 CLOSE ON 2023-01-01
        WHERE account ~ '(Income|Expenses):*'
        GROUP BY account, account_sortkey(account)
        ORDER BY account_sortkey(account);
        """
        self.assertRegex(
            out, 'Expenses:Taxes:401k *4.00 IRAUSD')

    @runshell
    def test_journal(self, out, err):
        """
        JOURNAL 'Assets:Checking'
        FROM OPEN ON 2022-02-01 CLOSE ON 2022-03-01;
        """
        self.assertRegex(out, "2022-01-31 +S +Opening balance for 'Assets:Checking'")
        self.assertRegex(out, "Test 02")

    @runshell
    def test_conversions(self, out, err):
        """
        SELECT date, payee, narration, position, balance
        FROM OPEN ON 2022-01-01 CLOSE ON 2023-01-01
        WHERE flag = 'C'
        """
        self.assertRegex(out, "2022-12-31 *Conversion for")

    @runshell
    def test_documents(self, out, err):
        """
        SELECT date, account, narration
        WHERE type = 'Document';
        """
        ## FIXME: Make this possible, we need an example with document entries.

    @runshell
    def test_holdings(self, out, err):
        """
        SELECT account, currency, cost_currency, sum(position)
        GROUP BY account, currency, cost_currency;
        """
        ## FIXME: Here we need to finally support FLATTEN to make this happen properly.


class TestRun(unittest.TestCase):

    @runshell
    def test_run_custom__list(self, out, err):
        """
        .run
        """
        self.assertEqual("home taxes",
                         re.sub(r'[] \n\t]+', ' ', out).strip())

    @runshell
    def test_run_custom__query_not_exists(self, out, err):
        """
        .run something
        """
        self.assertEqual('error: query "something" not found', err.strip())

    @runshell
    def test_run_custom__query_id(self, out, err):
        """
        .run taxes
        """
        self.assertRegex(out, 'date +description +position +balance')
        self.assertRegex(out, r'ACME \| Salary')

    @runshell
    def test_run_custom__query_string(self, out, err):
        """
        RUN "taxes"
        """
        self.assertRegex(out, 'date +description +position +balance')
        self.assertRegex(out, r'ACME \| Salary')

    @runshell
    def test_run_custom__all(self, out, err):
        """
        RUN *
        """
        self.assertRegex(out, 'date +description +position +balance')
        self.assertRegex(out, r'ACME \| Salary')
        self.assertRegex(out, 'account +total')
        self.assertRegex(out, 'Expenses:Home:Rent')


class TestCommands(unittest.TestCase):

    @runshell
    def test_tables(self, out, err):
        """
        .tables
        """
        self.assertEqual(out, textwrap.dedent('''\
            accounts
            balances
            commodities
            documents
            entries
            events
            notes
            postings
            prices
            transactions
        '''))


class TestHelp(unittest.TestCase):

    def test_help_functions(self):
        for name in dir(shell.BQLShell):
            if name.startswith('help_'):
                run_shell_command('help ' + name[5:])


class ClickTestCase(unittest.TestCase):
    """Base class for command-line program test cases."""

    def main(self, *args):
        init_filename = shell.INIT_FILENAME
        history_filename = shell.HISTORY_FILENAME
        try:
            shell.INIT_FILENAME = ''
            shell.HISTORY_FILENAME = ''
            runner = click.testing.CliRunner()
            result = runner.invoke(shell.main, args, catch_exceptions=False)
            self.assertEqual(result.exit_code, 0)
            return result
        finally:
            shell.INIT_FILENAME = init_filename
            shell.HISTORY_FILENAME = history_filename


class TestShell(ClickTestCase):

    @test_utils.docfile
    def test_success(self, filename):
        """
        2013-01-01 open Assets:Account1
        2013-01-01 open Assets:Account2
        2013-01-01 open Assets:Account3
        2013-01-01 open Equity:Unknown

        2013-04-05 *
          Equity:Unknown
          Assets:Account1     5000 USD

        2013-04-05 *
          Assets:Account1     -3000 USD
          Assets:Account2        30 BOOG {100 USD}

        2013-04-05 *
          Assets:Account1     -1000 USD
          Assets:Account3       800 EUR @ 1.25 USD
        """
        result = self.main(filename, "SELECT 1;")
        self.assertTrue(result.stdout)

    @test_utils.docfile
    def test_format_csv(self, filename):
        """
        """
        r = self.main(filename, '--format=csv', "SELECT 111 AS one, 222 AS two FROM #")
        self.assertEqual(r.stdout, textwrap.dedent('''\
            one,two
            111,222
        '''))

    @test_utils.docfile
    def test_format_text(self, filename):
        """
        """
        r = self.main(filename, '--format=text', "SELECT 111 AS one, 222 AS two FROM #")
        self.assertEqual(r.stdout, textwrap.dedent('''\
            one  two
            ---  ---
            111  222
        '''))


if __name__ == '__main__':
    unittest.main()



class Table:
    columns = {}
    name = ''

    def __getitem__(self, name):
        return self.columns[name]

    @property
    def wildcard_columns(self):
        # For backward compatibility. Remove once the postings table
        # is updated to return all columns upon ``SELECT *`` and the
        # query compiler is updated not to rely on this property
        return self.columns.keys()


class NullTable(Table):
    def __iter__(self):
        return iter([None])



from beanquery import tables
from beanquery.sources.beancount import GetItemColumn


class Table(tables.Table):
    def __init__(self, name, columns):
        self.name = name
        self.data = []
        self.columns = {}
        for cname, ctype in columns:
            self.columns[cname] = GetItemColumn(len(self.columns), ctype)

    def __iter__(self):
        return iter(self.data)

    def insert(self, row):
        assert len(row) == len(self.columns)
        self.data.append(row)


def create(name, columns, *args, **kwargs):
    return Table(name, columns)



#!/usr/bin/env python3

# WARNING: CAVEAT UTILITOR
#
#  This file was automatically generated by TatSu.
#
#     https://pypi.python.org/pypi/tatsu/
#
#  Any changes you make to it will be overwritten the next time
#  the file is generated.

# ruff: noqa: C405, COM812, I001, F401, PLR1702, PLC2801, SIM117

from __future__ import annotations

import sys
from pathlib import Path

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main


KEYWORDS: set[str] = {
    'AND',
    'AS',
    'ASC',
    'BY',
    'DESC',
    'DISTINCT',
    'FALSE',
    'FROM',
    'GROUP',
    'HAVING',
    'IN',
    'IS',
    'LIMIT',
    'NOT',
    'OR',
    'ORDER',
    'PIVOT',
    'SELECT',
    'TRUE',
    'WHERE',
    'CREATE',
    'TABLE',
    'USING',
    'INSERT',
    'INTO',
    'BALANCES',
    'JOURNAL',
    'PRINT',
}


class BQLBuffer(Buffer):
    def __init__(self, text, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=None,
            nameguard=None,
            ignorecase=True,
            namechars='',
            parseinfo=True,
            comments='(\\/\\*([^*]|[\\r\\n]|(\\*+([^*\\/]|[\\r\\n])))*\\*+\\/)',
            eol_comments='\\;[^\\n]*?$',
            keywords=KEYWORDS,
            start='bql',
        )
        config = config.replace(**settings)

        super().__init__(text, config=config)


class BQLParser(Parser):
    def __init__(self, /, config: ParserConfig | None = None, **settings):
        config = ParserConfig.new(
            config,
            owner=self,
            whitespace=None,
            nameguard=None,
            ignorecase=True,
            namechars='',
            parseinfo=True,
            comments='(\\/\\*([^*]|[\\r\\n]|(\\*+([^*\\/]|[\\r\\n])))*\\*+\\/)',
            eol_comments='\\;[^\\n]*?$',
            keywords=KEYWORDS,
            start='bql',
        )
        config = config.replace(**settings)

        super().__init__(config=config)

    @tatsumasu()
    def _bql_(self):
        self._statement_()
        self.name_last_node('@')
        with self._optional():
            self._token(';')
        self._check_eof()

    @tatsumasu()
    def _statement_(self):
        with self._choice():
            with self._option():
                self._select_()
            with self._option():
                self._balances_()
            with self._option():
                self._journal_()
            with self._option():
                self._print_()
            with self._option():
                self._create_table_()
            with self._option():
                self._insert_()
            self._error(
                'expecting one of: '
                "'BALANCES' 'CREATE' 'INSERT' 'JOURNAL'"
                "'PRINT' 'SELECT' <balances>"
                '<create_table> <insert> <journal>'
                '<print> <select>'
            )

    @tatsumasu('Select')
    def _select_(self):
        self._token('SELECT')
        with self._optional():
            self._token('DISTINCT')
            self._constant(True)
            self.name_last_node('distinct')
            self._define(['distinct'], [])
        with self._group():
            with self._choice():
                with self._option():

                    def sep0():
                        self._token(',')

                    def block1():
                        self._target_()
                    self._positive_gather(block1, sep0)
                with self._option():
                    self._asterisk_()
                self._error(
                    'expecting one of: '
                    '<asterisk> <target>'
                )
        self.name_last_node('targets')
        with self._optional():
            self._token('FROM')
            with self._group():
                with self._choice():
                    with self._option():
                        self.__table_()
                    with self._option():
                        self._subselect_()
                    with self._option():
                        self._from_()
                    self._error(
                        'expecting one of: '
                        '<_table> <from> <subselect>'
                    )
            self.name_last_node('from_clause')
            self._define(['from_clause'], [])
        with self._optional():
            self._token('WHERE')
            self._expression_()
            self.name_last_node('where_clause')
            self._define(['where_clause'], [])
        with self._optional():
            self._token('GROUP')
            self._token('BY')
            self._groupby_()
            self.name_last_node('group_by')
            self._define(['group_by'], [])
        with self._optional():
            self._token('ORDER')
            self._token('BY')

            def sep2():
                self._token(',')

            def block3():
                self._order_()
            self._positive_gather(block3, sep2)
            self.name_last_node('order_by')
            self._define(['order_by'], [])
        with self._optional():
            self._token('PIVOT')
            self._token('BY')
            self._pivotby_()
            self.name_last_node('pivot_by')
            self._define(['pivot_by'], [])
        with self._optional():
            self._token('LIMIT')
            self._integer_()
            self.name_last_node('limit')
            self._define(['limit'], [])
        self._define(['distinct', 'from_clause', 'group_by', 'limit', 'order_by', 'pivot_by', 'targets', 'where_clause'], [])

    @tatsumasu()
    def _subselect_(self):
        self._token('(')
        self._select_()
        self.name_last_node('@')
        self._token(')')

    @tatsumasu('From')
    @nomemo
    def _from_(self):
        with self._choice():
            with self._option():
                self._token('OPEN')
                self._cut()
                self._token('ON')
                self._date_()
                self.name_last_node('open')
                with self._optional():
                    self._token('CLOSE')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('ON')
                                self._date_()
                                self.name_last_node('close')
                                self._define(['close'], [])
                            with self._option():
                                self._empty_closure()
                                self._constant(True)
                                self.name_last_node('close')
                                self._define(['close'], [])
                            self._error(
                                'expecting one of: '
                                "'ON'"
                            )
                    self._define(['close'], [])
                with self._optional():
                    self._token('CLEAR')
                    self._constant(True)
                    self.name_last_node('clear')
                    self._define(['clear'], [])
                self._define(['clear', 'close', 'open'], [])
            with self._option():
                self._token('CLOSE')
                self._cut()
                with self._group():
                    with self._choice():
                        with self._option():
                            self._token('ON')
                            self._date_()
                            self.name_last_node('close')
                            self._define(['close'], [])
                        with self._option():
                            self._empty_closure()
                            self._constant(True)
                            self.name_last_node('close')
                            self._define(['close'], [])
                        self._error(
                            'expecting one of: '
                            "'ON'"
                        )
                with self._optional():
                    self._token('CLEAR')
                    self._constant(True)
                    self.name_last_node('clear')
                    self._define(['clear'], [])
                self._define(['clear', 'close'], [])
            with self._option():
                self._token('CLEAR')
                self._cut()
                self._constant(True)
                self.name_last_node('clear')
                self._define(['clear'], [])
            with self._option():
                self._expression_()
                self.name_last_node('expression')
                with self._optional():
                    self._token('OPEN')
                    self._token('ON')
                    self._date_()
                    self.name_last_node('open')
                    self._define(['open'], [])
                with self._optional():
                    self._token('CLOSE')
                    with self._group():
                        with self._choice():
                            with self._option():
                                self._token('ON')
                                self._date_()
                                self.name_last_node('close')
                                self._define(['close'], [])
                            with self._option():
                                self._empty_closure()
                                self._constant(True)
                                self.name_last_node('close')
                                self._define(['close'], [])
                            self._error(
                                'expecting one of: '
                                "'ON'"
                            )
                    self._define(['close'], [])
                with self._optional():
                    self._token('CLEAR')
                    self._constant(True)
                    self.name_last_node('clear')
                    self._define(['clear'], [])
                self._define(['clear', 'close', 'expression', 'open'], [])
            self._error(
                'expecting one of: '
                "'CLEAR' 'CLOSE' 'OPEN' <conjunction>"
                '<disjunction> <expression>'
            )

    @tatsumasu('Table')
    def __table_(self):
        with self._choice():
            with self._option():
                self._pattern('#([a-zA-Z_][a-zA-Z0-9_]*)?')
                self.name_last_node('name')
            with self._option():
                self._quoted_identifier_()
                self.name_last_node('name')
            self._error(
                'expecting one of: '
                '#([a-zA-Z_][a-zA-Z0-9_]*)?'
                '<quoted_identifier>'
                '\\"((?:[^\\"]|\\"\\")+)\\"'
            )

    @tatsumasu('Table')
    def _table_(self):
        self._identifier_()
        self.name_last_node('name')

    @tatsumasu('GroupBy')
    def _groupby_(self):

        def sep0():
            self._token(',')

        def block1():
            with self._group():
                with self._choice():
                    with self._option():
                        self._integer_()
                    with self._option():
                        self._expression_()
                    self._error(
                        'expecting one of: '
                        '<expression> <integer>'
                    )
        self._positive_gather(block1, sep0)
        self.name_last_node('columns')
        with self._optional():
            self._token('HAVING')
            self._expression_()
            self.name_last_node('having')
            self._define(['having'], [])
        self._define(['columns', 'having'], [])

    @tatsumasu('OrderBy')
    def _order_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._integer_()
                with self._option():
                    self._expression_()
                self._error(
                    'expecting one of: '
                    '<expression> <integer>'
                )
        self.name_last_node('column')
        self._ordering_()
        self.name_last_node('ordering')
        self._define(['column', 'ordering'], [])

    @tatsumasu()
    def _ordering_(self):
        with self._optional():
            with self._choice():
                with self._option():
                    self._token('DESC')
                with self._option():
                    self._token('ASC')
                self._error(
                    'expecting one of: '
                    "'ASC' 'DESC'"
                )

    @tatsumasu('PivotBy')
    def _pivotby_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._integer_()
                with self._option():
                    self._column_()
                self._error(
                    'expecting one of: '
                    '<column> <integer>'
                )
        self.add_last_node_to_name('columns')
        self._token(',')
        with self._group():
            with self._choice():
                with self._option():
                    self._integer_()
                with self._option():
                    self._column_()
                self._error(
                    'expecting one of: '
                    '<column> <integer>'
                )
        self.add_last_node_to_name('columns')
        self._define(
            [],
            ['columns'],
        )

    @tatsumasu('Target')
    def _target_(self):
        self._expression_()
        self.name_last_node('expression')
        with self._optional():
            self._token('AS')
            self._identifier_()
            self.name_last_node('name')
            self._define(['name'], [])
        self._define(['expression', 'name'], [])

    @tatsumasu()
    @nomemo
    def _expression_(self):
        with self._choice():
            with self._option():
                self._disjunction_()
            with self._option():
                self._conjunction_()
            self._error(
                'expecting one of: '
                '<and> <conjunction> <disjunction>'
                '<inversion> <or>'
            )

    @tatsumasu()
    @nomemo
    def _disjunction_(self):
        with self._choice():
            with self._option():
                self._or_()
            with self._option():
                self._conjunction_()
            self._error(
                'expecting one of: '
                '<and> <conjunction> <inversion> <or>'
            )

    @tatsumasu('Or')
    @nomemo
    def _or_(self):
        self._conjunction_()
        self.add_last_node_to_name('args')

        def block0():
            self._token('OR')
            self._conjunction_()
            self.add_last_node_to_name('args')
            self._define(
                [],
                ['args'],
            )
        self._positive_closure(block0)
        self._define(
            [],
            ['args'],
        )

    @tatsumasu()
    @nomemo
    def _conjunction_(self):
        with self._choice():
            with self._option():
                self._and_()
            with self._option():
                self._inversion_()
            self._error(
                'expecting one of: '
                '<and> <comparison> <inversion> <not>'
            )

    @tatsumasu('And')
    @nomemo
    def _and_(self):
        self._inversion_()
        self.add_last_node_to_name('args')

        def block0():
            self._token('AND')
            self._inversion_()
            self.add_last_node_to_name('args')
            self._define(
                [],
                ['args'],
            )
        self._positive_closure(block0)
        self._define(
            [],
            ['args'],
        )

    @tatsumasu()
    @nomemo
    def _inversion_(self):
        with self._choice():
            with self._option():
                self._not_()
            with self._option():
                self._comparison_()
            self._error(
                'expecting one of: '
                "'NOT' <all> <any> <between> <comparison>"
                '<eq> <gt> <gte> <in> <isnotnull>'
                '<isnull> <lt> <lte> <match> <matches>'
                '<neq> <not> <notin> <notmatch> <sum>'
            )

    @tatsumasu('Not')
    def _not_(self):
        self._token('NOT')
        self._inversion_()
        self.name_last_node('operand')
        self._define(['operand'], [])

    @tatsumasu()
    @nomemo
    def _comparison_(self):
        with self._choice():
            with self._option():
                self._any_()
            with self._option():
                self._all_()
            with self._option():
                self._lt_()
            with self._option():
                self._lte_()
            with self._option():
                self._gt_()
            with self._option():
                self._gte_()
            with self._option():
                self._eq_()
            with self._option():
                self._neq_()
            with self._option():
                self._in_()
            with self._option():
                self._notin_()
            with self._option():
                self._match_()
            with self._option():
                self._notmatch_()
            with self._option():
                self._matches_()
            with self._option():
                self._isnull_()
            with self._option():
                self._isnotnull_()
            with self._option():
                self._between_()
            with self._option():
                self._sum_()
            self._error(
                'expecting one of: '
                '<add> <all> <any> <between> <eq> <gt>'
                '<gte> <in> <isnotnull> <isnull> <lt>'
                '<lte> <match> <matches> <neq> <notin>'
                '<notmatch> <sub> <sum> <term>'
            )

    @tatsumasu('Any')
    @nomemo
    def _any_(self):
        self._sum_()
        self.name_last_node('left')
        self._op_()
        self.name_last_node('op')
        self._token('any')
        self._token('(')
        self._expression_()
        self.name_last_node('right')
        self._token(')')
        self._define(['left', 'op', 'right'], [])

    @tatsumasu('All')
    def _all_(self):
        self._sum_()
        self.name_last_node('left')
        self._op_()
        self.name_last_node('op')
        self._token('all')
        self._token('(')
        self._expression_()
        self.name_last_node('right')
        self._token(')')
        self._define(['left', 'op', 'right'], [])

    @tatsumasu()
    def _op_(self):
        with self._choice():
            with self._option():
                self._token('<')
            with self._option():
                self._token('<=')
            with self._option():
                self._token('>')
            with self._option():
                self._token('>=')
            with self._option():
                self._token('=')
            with self._option():
                self._token('!=')
            with self._option():
                self._token('~')
            with self._option():
                self._token('!~')
            with self._option():
                self._token('?~')
            self._error(
                'expecting one of: '
                "'!=' '!~' '<' '<=' '=' '>' '>=' '?~' '~'"
            )

    @tatsumasu('Less')
    def _lt_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('<')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('LessEq')
    def _lte_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('<=')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Greater')
    def _gt_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('>')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('GreaterEq')
    def _gte_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('>=')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Equal')
    def _eq_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('=')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('NotEqual')
    def _neq_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('!=')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('In')
    def _in_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('IN')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('NotIn')
    def _notin_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('NOT')
        self._token('IN')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Match')
    def _match_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('~')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('NotMatch')
    def _notmatch_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('!~')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Matches')
    def _matches_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('?~')
        self._sum_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('IsNull')
    def _isnull_(self):
        self._sum_()
        self.name_last_node('operand')
        self._token('IS')
        self._token('NULL')
        self._define(['operand'], [])

    @tatsumasu('IsNotNull')
    def _isnotnull_(self):
        self._sum_()
        self.name_last_node('operand')
        self._token('IS')
        self._token('NOT')
        self._token('NULL')
        self._define(['operand'], [])

    @tatsumasu('Between')
    def _between_(self):
        self._sum_()
        self.name_last_node('operand')
        self._token('BETWEEN')
        self._sum_()
        self.name_last_node('lower')
        self._token('AND')
        self._sum_()
        self.name_last_node('upper')
        self._define(['lower', 'operand', 'upper'], [])

    @tatsumasu()
    @leftrec
    def _sum_(self):
        with self._choice():
            with self._option():
                self._add_()
            with self._option():
                self._sub_()
            with self._option():
                self._term_()
            self._error(
                'expecting one of: '
                '<add> <div> <factor> <mod> <mul> <sub>'
                '<sum> <term>'
            )

    @tatsumasu('Add')
    @nomemo
    def _add_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('+')
        self._cut()
        self._term_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Sub')
    @nomemo
    def _sub_(self):
        self._sum_()
        self.name_last_node('left')
        self._token('-')
        self._cut()
        self._term_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu()
    @leftrec
    def _term_(self):
        with self._choice():
            with self._option():
                self._mul_()
            with self._option():
                self._div_()
            with self._option():
                self._mod_()
            with self._option():
                self._factor_()
            self._error(
                'expecting one of: '
                "'(' <div> <factor> <mod> <mul> <term>"
                '<unary>'
            )

    @tatsumasu('Mul')
    @nomemo
    def _mul_(self):
        self._term_()
        self.name_last_node('left')
        self._token('*')
        self._cut()
        self._factor_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Div')
    @nomemo
    def _div_(self):
        self._term_()
        self.name_last_node('left')
        self._token('/')
        self._cut()
        self._factor_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu('Mod')
    @nomemo
    def _mod_(self):
        self._term_()
        self.name_last_node('left')
        self._token('%')
        self._cut()
        self._factor_()
        self.name_last_node('right')
        self._define(['left', 'right'], [])

    @tatsumasu()
    @nomemo
    def _factor_(self):
        with self._choice():
            with self._option():
                self._unary_()
            with self._option():
                self._token('(')
                self._expression_()
                self.name_last_node('@')
                self._token(')')
            self._error(
                'expecting one of: '
                "'(' <primary> <uminus> <unary> <uplus>"
            )

    @tatsumasu()
    @nomemo
    def _unary_(self):
        with self._choice():
            with self._option():
                self._uplus_()
            with self._option():
                self._uminus_()
            with self._option():
                self._primary_()
            self._error(
                'expecting one of: '
                "'+' '-' <atom> <attribute> <primary>"
                '<subscript> <uminus> <uplus>'
            )

    @tatsumasu()
    def _uplus_(self):
        self._token('+')
        self._atom_()
        self.name_last_node('@')

    @tatsumasu('Neg')
    def _uminus_(self):
        self._token('-')
        self._factor_()
        self.name_last_node('operand')
        self._define(['operand'], [])

    @tatsumasu()
    @leftrec
    def _primary_(self):
        with self._choice():
            with self._option():
                self._attribute_()
            with self._option():
                self._subscript_()
            with self._option():
                self._atom_()
            self._error(
                'expecting one of: '
                "'SELECT' <atom> <attribute> <column>"
                '<constant> <function> <placeholder>'
                '<primary> <select> <subscript>'
            )

    @tatsumasu('Attribute')
    @nomemo
    def _attribute_(self):
        self._primary_()
        self.name_last_node('operand')
        self._token('.')
        self._identifier_()
        self.name_last_node('name')
        self._define(['name', 'operand'], [])

    @tatsumasu('Subscript')
    @nomemo
    def _subscript_(self):
        self._primary_()
        self.name_last_node('operand')
        self._token('[')
        self._string_()
        self.name_last_node('key')
        self._token(']')
        self._define(['key', 'operand'], [])

    @tatsumasu()
    def _atom_(self):
        with self._choice():
            with self._option():
                self._select_()
            with self._option():
                self._function_()
            with self._option():
                self._constant_()
            with self._option():
                self._column_()
            with self._option():
                self._placeholder_()
            self._error(
                'expecting one of: '
                "'%(' '%s' 'SELECT' <boolean> <column>"
                '<constant> <date> <decimal> <function>'
                '<identifier> <integer> <list> <literal>'
                '<null> <placeholder> <select> <string>'
            )

    @tatsumasu('Placeholder')
    def _placeholder_(self):
        with self._choice():
            with self._option():
                self._token('%s')
                self._constant('')
                self.name_last_node('name')
                self._define(['name'], [])
            with self._option():
                self._token('%(')
                self._identifier_()
                self.name_last_node('name')
                self._token(')s')
                self._define(['name'], [])
            self._error(
                'expecting one of: '
                "'%(' '%s'"
            )

    @tatsumasu('Function')
    def _function_(self):
        with self._choice():
            with self._option():
                self._identifier_()
                self.name_last_node('fname')
                self._token('(')

                def sep0():
                    self._token(',')

                def block1():
                    self._expression_()
                self._gather(block1, sep0)
                self.name_last_node('operands')
                self._token(')')
                self._define(['fname', 'operands'], [])
            with self._option():
                self._identifier_()
                self.name_last_node('fname')
                self._token('(')
                self._asterisk_()
                self.add_last_node_to_name('operands')
                self._token(')')
                self._define(
                    ['fname'],
                    ['operands'],
                )
            self._error(
                'expecting one of: '
                '<identifier> <quoted_identifier>'
                '<unquoted_identifier>'
            )

    @tatsumasu('Column')
    def _column_(self):
        self._identifier_()
        self.name_last_node('name')

    @tatsumasu()
    def _literal_(self):
        with self._choice():
            with self._option():
                self._date_()
            with self._option():
                self._decimal_()
            with self._option():
                self._integer_()
            with self._option():
                self._string_()
            with self._option():
                self._null_()
            with self._option():
                self._boolean_()
            self._error(
                'expecting one of: '
                "'FALSE' 'NULL' 'TRUE' <boolean> <date>"
                '<decimal> <integer> <null> <string>'
                '[0-9]+ [0-9]+\\.[0-9]*|[0-9]*\\.[0-9]+'
                '[0-9]{4}-[0-9]{2}-[0-9]{2}'
                '\\"[^\\"]*\\"|\\\'(?:[^\\\']|\\\'\\\')*\\\''
            )

    @tatsumasu('Constant')
    def _constant_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._literal_()
                with self._option():
                    self._list_()
                self._error(
                    'expecting one of: '
                    "'(' 'FALSE' 'NULL' 'TRUE' <boolean>"
                    '<date> <decimal> <integer> <list>'
                    '<literal> <null> <string> [0-9]+'
                    '[0-9]+\\.[0-9]*|[0-9]*\\.[0-9]+'
                    '[0-9]{4}-[0-9]{2}-[0-9]{2}'
                    '\\"[^\\"]*\\"|\\\'(?:[^\\\']|\\\'\\\')*\\\''
                )
        self.name_last_node('value')

    @tatsumasu()
    def _list_(self):
        self._token('(')
        with self._if():
            with self._group():
                self._literal_()
                self._token(',')

        def sep0():
            self._token(',')

        def block1():
            with self._group():
                with self._choice():
                    with self._option():
                        self._literal_()
                    with self._option():
                        self._void()
                    self._error(
                        'expecting one of: '
                        '<literal>'
                    )
        self._positive_gather(block1, sep0)
        self.name_last_node('@')
        self._token(')')

    @tatsumasu()
    def _identifier_(self):
        with self._choice():
            with self._option():
                self._unquoted_identifier_()
            with self._option():
                self._quoted_identifier_()
            self._error(
                'expecting one of: '
                '<quoted_identifier>'
                '<unquoted_identifier> [a-zA-'
                'Z_][a-zA-Z0-9_]* \\"((?:[^\\"]|\\"\\")+)\\"'
            )

    @tatsumasu()
    @isname
    def _unquoted_identifier_(self):
        self._pattern('[a-zA-Z_][a-zA-Z0-9_]*')

    @tatsumasu()
    def _quoted_identifier_(self):
        self._pattern('\\"((?:[^\\"]|\\"\\")+)\\"')

    @tatsumasu()
    def _asterisk_(self):
        self._token('*')

    @tatsumasu()
    def _string_(self):
        self._pattern('\\"[^\\"]*\\"|\\\'(?:[^\\\']|\\\'\\\')*\\\'')

    @tatsumasu()
    def _boolean_(self):
        with self._choice():
            with self._option():
                self._token('TRUE')
            with self._option():
                self._token('FALSE')
            self._error(
                'expecting one of: '
                "'FALSE' 'TRUE'"
            )

    @tatsumasu()
    def _null_(self):
        self._token('NULL')

    @tatsumasu()
    def _integer_(self):
        self._pattern('[0-9]+')

    @tatsumasu()
    def _decimal_(self):
        self._pattern('[0-9]+\\.[0-9]*|[0-9]*\\.[0-9]+')

    @tatsumasu()
    def _date_(self):
        self._pattern('[0-9]{4}-[0-9]{2}-[0-9]{2}')

    @tatsumasu('Balances')
    def _balances_(self):
        self._token('BALANCES')
        with self._optional():
            self._token('AT')
            self._identifier_()
            self.name_last_node('summary_func')
            self._define(['summary_func'], [])
        with self._optional():
            self._token('FROM')
            self._from_()
            self.name_last_node('from_clause')
            self._define(['from_clause'], [])
        with self._optional():
            self._token('WHERE')
            self._expression_()
            self.name_last_node('where_clause')
            self._define(['where_clause'], [])
        self._define(['from_clause', 'summary_func', 'where_clause'], [])

    @tatsumasu('Journal')
    def _journal_(self):
        self._token('JOURNAL')
        with self._optional():
            self._string_()
            self.name_last_node('account')
        with self._optional():
            self._token('AT')
            self._identifier_()
            self.name_last_node('summary_func')
            self._define(['summary_func'], [])
        with self._optional():
            self._token('FROM')
            self._from_()
            self.name_last_node('from_clause')
            self._define(['from_clause'], [])
        self._define(['account', 'from_clause', 'summary_func'], [])

    @tatsumasu('Print')
    def _print_(self):
        self._token('PRINT')
        with self._optional():
            self._token('FROM')
            self._from_()
            self.name_last_node('from_clause')
            self._define(['from_clause'], [])
        self._define(['from_clause'], [])

    @tatsumasu('CreateTable')
    def _create_table_(self):
        self._token('CREATE')
        self._token('TABLE')
        self._cut()
        self._identifier_()
        self.name_last_node('name')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('(')

                    def sep0():
                        self._token(',')

                    def block1():
                        with self._group():
                            self._identifier_()
                            self._identifier_()
                    self._gather(block1, sep0)
                    self.name_last_node('columns')
                    self._token(')')
                    with self._optional():
                        self._token('USING')
                        self._string_()
                        self.name_last_node('using')
                        self._define(['using'], [])
                    self._define(['columns', 'using'], [])
                with self._option():
                    self._token('USING')
                    self._string_()
                    self.name_last_node('using')
                    self._define(['using'], [])
                with self._option():
                    self._token('AS')
                    self._select_()
                    self.name_last_node('query')
                    self._define(['query'], [])
                self._error(
                    'expecting one of: '
                    "'(' 'AS' 'USING'"
                )
        self._define(['columns', 'name', 'query', 'using'], [])

    @tatsumasu('Insert')
    def _insert_(self):
        self._token('INSERT')
        self._token('INTO')
        self._cut()
        self._table_()
        self.name_last_node('table')
        with self._optional():
            self._token('(')

            def sep0():
                self._token(',')

            def block1():
                self._column_()
            self._gather(block1, sep0)
            self.name_last_node('columns')
            self._token(')')
            self._define(['columns'], [])
        self._token('VALUES')
        self._token('(')

        def sep2():
            self._token(',')

        def block3():
            self._expression_()
        self._gather(block3, sep2)
        self.name_last_node('values')
        self._token(')')
        self._define(['columns', 'table', 'values'], [])


def main(filename, **kwargs):
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        text = Path(filename).read_text()
    parser = BQLParser()
    return parser.parse(
        text,
        filename=filename,
        **kwargs,
    )


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, BQLParser, name='BQL')
    data = asjson(ast)
    print(json.dumps(data, indent=2))



import re
from urllib.parse import urlparse, parse_qsl

from beanquery import tables
from beanquery.query_compile import EvalColumn


class Column(EvalColumn):
    def __init__(self, datatype, func):
        super().__init__(datatype)
        self.func = func

    def __call__(self, row):
        return self.func(row)


class TableMeta(type):
    def __new__(mcs, name, bases, dct):
        columns = {}
        members = {}
        for key, value in dct.items():
            if isinstance(value, EvalColumn):
                value.name = key
                columns[key] = value
                continue
            members[key] = value
        assert 'columns' not in members
        members['columns'] = columns
        return super().__new__(mcs, name, bases, members)


class Table(tables.Table, metaclass=TableMeta):
    x = Column(int, lambda row: row)

    def __init__(self, *args):
        self.rows = range(*args)

    def __iter__(self):
        return iter(self.rows)


class MagicColumnsRegistry(dict):
    def get(self, key, default=None):
        if re.fullmatch(r'x+', key):
            return Column(int, lambda x: x * len(key))
        return default


class MagicTable(tables.Table):
    columns = MagicColumnsRegistry()

    def __init__(self, *args):
        self.rows = range(*args)

    def __iter__(self):
        return iter(self.rows)


def create(name, columns, using):
    assert columns is None
    parts = urlparse(using)
    params = dict(parse_qsl(parts.query, strict_parsing=True))
    cls = MagicTable if parts.path == 'magic' else Table
    return cls(int(params.get('start', 0)), int(params.get('stop', 11)), int(params.get('step', 1)))


def attach(context, dsn):
    parts = urlparse(dsn)
    params = dict(parse_qsl(parts.query, strict_parsing=True))
    name = params.get('name', 'test')
    context.tables[name] = create(name, None, dsn)



"""Execution of interpreter on data rows.
"""
__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import itertools
import operator

from . import compiler
from . import hashable
from . import cursor



class Unique:
    def __init__(self, columns):
        self.wrap = hashable.make(columns)

    def __call__(self, iterable):
        wrap = self.wrap
        seen = set()
        add = seen.add
        for obj in iterable:
            h = wrap(obj)
            if h not in seen:
                add(h)
                yield obj


class Allocator:
    """A helper class to count slot allocations and return unique handles to them.
    """
    def __init__(self):
        self.size = 0

    def allocate(self):
        """Allocate a new slot to store row aggregation information.

        Returns:
          A unique handle used to index into an row-aggregation store (an integer).
        """
        handle = self.size
        self.size += 1
        return handle

    def create_store(self):
        """Create a new row-aggregation store suitable to contain all the node allocations.

        Returns:
          A store that can accommodate and be indexed by all the allocated slot handles.
        """
        return [None] * self.size


class NullType:
    """An object that compares smaller than anything.

    An instance of this class is used to replace None in BQL query
    results in sort keys to obtain sorting semantics similar to SQL
    where NULL is sortet at the beginning.

    """
    __slots__ = ()

    def __repr__(self):
        return 'NULL'

    __str__ = __repr__

    def __lt__(self, other):
        # Make sure that instances of this class compare equal.
        if isinstance(other, NullType):
            return False
        return True

    def __gt__(self, other):
        # Make sure that instances of this class compare equal.
        if isinstance(other, NullType):
            return True
        return False


NULL = NullType()


def nullitemgetter(item, *items):
    """An itemgetter() that replaces None values with NULL."""
    if items:
        items = (item, *items)
        def func(obj):
            r = []
            for i in items:
                value = obj[i]
                r.append(value if value is not None else NULL)
            return tuple(r)
        return func
    def func(obj):
        value = obj[item]
        return value if value is not None else NULL
    return func


def execute_select(query):
    """Given a compiled select statement, execute the query.

    Args:
      query: An instance of a query_compile.Query
      entries: A list of directives.
      options_map: A parser's option_map.
    Returns:
      A pair of:
        result_types: A list of (name, data-type) item pairs.
        result_rows: A list of ResultRow tuples of length and types described by
          'result_types'.
    """
    # Figure out the result types that describe what we return.
    result_types = tuple(cursor.Column(target.name, target.c_expr.dtype)
                         for target in query.c_targets
                         if target.name is not None)

    # Pre-compute lists of the expressions to evaluate.
    group_indexes = (set(query.group_indexes)
                     if query.group_indexes is not None
                     else query.group_indexes)

    # Indexes of the columns for result rows and order rows.
    result_indexes = [index
                      for index, c_target in enumerate(query.c_targets)
                      if c_target.name]
    order_spec = query.order_spec

    # Dispatch between the non-aggregated queries and aggregated queries.
    c_where = query.c_where
    rows = []

    # Precompute a list of expressions to be evaluated.
    c_target_exprs = [c_target.c_expr for c_target in query.c_targets]

    if query.group_indexes is None:
        # This is a non-aggregated query.

        # Iterate over all the postings once.
        for context in query.table:
            if c_where is None or c_where(context):
                values = [c_expr(context) for c_expr in c_target_exprs]
                rows.append(values)

    else:
        # This is an aggregated query.

        # Precompute lists of non-aggregate and aggregate expressions to
        # evaluate. For aggregate targets, we hunt down the aggregate
        # sub-expressions to evaluate, to avoid recursion during iteration.
        c_nonaggregate_exprs = []
        c_aggregate_exprs = []
        for index, c_expr in enumerate(c_target_exprs):
            if index in group_indexes:
                c_nonaggregate_exprs.append(c_expr)
            else:
                _, aggregate_exprs = compiler.get_columns_and_aggregates(c_expr)
                c_aggregate_exprs.extend(aggregate_exprs)
        # Note: it is possible that there are no aggregates to compute here. You could
        # have all columns be non-aggregates and group-by the entire list of columns.

        # Pre-allocate handles in aggregation nodes.
        allocator = Allocator()
        for c_expr in c_aggregate_exprs:
            c_expr.allocate(allocator)

        def create():
            # Create a new row in the aggregates store.
            store = allocator.create_store()
            for c_expr in c_aggregate_exprs:
                c_expr.initialize(store)
            return store

        context = None
        aggregates = collections.defaultdict(create)

        # Iterate over all the postings to evaluate the aggregates.
        for context in query.table:
            if c_where is None or c_where(context):

                # Compute the non-aggregate expressions.
                key = tuple(c_expr(context) for c_expr in c_nonaggregate_exprs)

                # Get an appropriate store for the unique key of this row.
                store = aggregates[key]

                # Update the aggregate expressions.
                for c_expr in c_aggregate_exprs:
                    c_expr.update(store, context)

        # Iterate over all the aggregations.
        for key, store in aggregates.items():
            key_iter = iter(key)
            values = []

            # Finalize the store.
            for c_expr in c_aggregate_exprs:
                c_expr.finalize(store)

            for index, c_expr in enumerate(c_target_exprs):
                if index in group_indexes:
                    value = next(key_iter)
                else:
                    value = c_expr(context)
                values.append(value)

            # Skip row if HAVING clause expression is false.
            if query.having_index is not None:
                if not values[query.having_index]:
                    continue

            rows.append(values)

    # Apply ORDER BY.
    if order_spec is not None:
        # Process the order-by clauses grouped by their ordering direction.
        for reverse, spec in itertools.groupby(reversed(order_spec), key=operator.itemgetter(1)):
            indexes = reversed([i[0] for i in spec])
            # The rows may contain None values: nullitemgetter()
            # replaces these with a special value that compares
            # smaller than anything else.
            rows.sort(key=nullitemgetter(*indexes), reverse=reverse)

    # Extract results set and convert into tuples.
    rows = (tuple(row[i] for i in result_indexes) for row in rows)

    # Apply DISTINCT.
    if query.distinct:
        unique = Unique(result_types)
        rows = unique(rows)

    # Apply LIMIT.
    if query.limit is not None:
        rows = itertools.islice(rows, query.limit)

    return result_types, list(rows)
